@@ -138,7 +138,7 @@ class CommonTsetlinMachine():
 		return X_transformed.reshape((number_of_examples, self.number_of_classes*self.number_of_clauses))
 	
 class MultiClassConvolutionalTsetlinMachine2D(CommonTsetlinMachine):
-	def __init__(self, number_of_clauses, T, s, patch_dim, boost_true_positive_feedback=1, number_of_state_bits=8, append_negated=True, max_weight=1):
+	def __init__(self, number_of_clauses, T, s, patch_dim, boost_true_positive_feedback=1, number_of_state_bits=8, append_negated=True, max_weight=1, number_of_classes=None):
 		self.number_of_clauses = number_of_clauses
 		self.number_of_clause_chunks = (number_of_clauses-1)/32 + 1
 		self.number_of_state_bits = number_of_state_bits
@@ -147,9 +147,9 @@ class MultiClassConvolutionalTsetlinMachine2D(CommonTsetlinMachine):
 		self.s = s
 		self.max_weight = min(max_weight,255)
 		self.boost_true_positive_feedback = boost_true_positive_feedback
-
 		self.append_negated = append_negated
-		
+		self.number_of_classes = number_of_classes
+
 		self.X_train = np.array([])
 		self.Y_train = np.array([])
 		self.X_test = np.array([])
@@ -167,7 +167,8 @@ class MultiClassConvolutionalTsetlinMachine2D(CommonTsetlinMachine):
 			self.X_train = X
 			self.Y_train = Y
 			
-			self.number_of_classes = int(np.max(Y) + 1)
+			if self.number_of_classes == None:
+				self.number_of_classes = int(np.max(Y) + 1)
 			
 			self.min_y = None
 			self.max_y = None
@@ -287,303 +288,69 @@ class MultiClassConvolutionalTsetlinMachine2D(CommonTsetlinMachine):
 		return np.argmax(self.score(X), axis=0)
 
 class MultiClassTsetlinMachine(CommonTsetlinMachine):
-	def __init__(self, number_of_clauses, T, s, boost_true_positive_feedback=1, number_of_state_bits=8, append_negated=True, max_weight=1):
+	def __init__(self, number_of_clauses, T, s, boost_true_positive_feedback=1, number_of_state_bits=8, append_negated=True, max_weight=1, number_of_classes = None):
 		self.number_of_clauses = number_of_clauses
-		self.number_of_clause_chunks = (number_of_clauses-1)/32 + 1
-		self.number_of_state_bits = number_of_state_bits
 		self.T = int(T)
 		self.s = s
-		self.max_weight = min(max_weight,255)
 		self.boost_true_positive_feedback = boost_true_positive_feedback
-		
+		self.number_of_state_bits = number_of_state_bits
 		self.append_negated = append_negated
+		self.max_weight = max_weight
+		self.number_of_classes = number_of_classes
 
 		self.X_train = np.array([])
 		self.Y_train = np.array([])
-		self.X_test = np.array([])
-		self.ta_state = np.array([])
-		self.clause_weights = np.array([])
-
-		mod_encode = SourceModule(kernels.code_encode, no_extern_c=True)
-		self.prepare_encode = mod_encode.get_function("prepare_encode")
-		self.encode = mod_encode.get_function("encode")
 
 	def fit(self, X, Y, epochs=100, incremental=False, batch_size = 100):
-		number_of_examples = X.shape[0]
-		number_of_example_chunks = int((number_of_examples-1)/32 + 1)
+		X = X.reshape(X.shape[0], X.shape[1], 1)
 
 		if (not np.array_equal(self.X_train, X)) or (not np.array_equal(self.Y_train, Y)):
 			self.X_train = X
 			self.Y_train = Y
 
-			self.number_of_classes = int(np.max(Y) + 1)
-
-			self.min_y = None
-			self.max_y = None
-			
-			self.dim = (X.shape[1], 1, 1)
 			self.patch_dim = (X.shape[1], 1, 1)
 
-			if self.append_negated:
-				self.number_of_features = X.shape[1]*2
-			else:
-				self.number_of_features = X.shape[1]
+			self.ctm = MultiClassConvolutionalTsetlinMachine2D(self.number_of_clauses, self.T, self.s, self.patch_dim, boost_true_positive_feedback=self.boost_true_positive_feedback, number_of_state_bits=self.number_of_state_bits, append_negated=self.append_negated, max_weight=self.max_weight, number_of_classes=self.number_of_classes)
 
-			self.number_of_patches = 1
-			self.number_of_ta_chunks = int((self.number_of_features-1)/32 + 1)
-			
-			parameters = """
-	#define CLASSES %d
-	#define CLAUSES %d
-	#define FEATURES %d
-	#define STATE_BITS %d
-	#define BOOST_TRUE_POSITIVE_FEEDBACK %d
-	#define S %f
-	#define THRESHOLD %d
-	#define MAX_WEIGHT %d
-
-	#define NEGATIVE_CLAUSES %d
-
-	#define PATCHES %d
-
-	#define NUMBER_OF_EXAMPLES %d
-
-	#define BATCH_SIZE %d
-
-""" % (self.number_of_classes, self.number_of_clauses, self.number_of_features, self.number_of_state_bits, int(self.boost_true_positive_feedback), self.s, self.T, self.max_weight, 1, 1, number_of_examples, batch_size)
-
-			mod_prepare = SourceModule(parameters + kernels.code_header + kernels.code_prepare, no_extern_c=True)
-			self.prepare = mod_prepare.get_function("prepare")
-
-			self.allocate_gpu_memory(number_of_examples)
-
-			self.prepare(self.ta_state_gpu, self.clause_weights_gpu, self.clause_output_gpu, self.class_sum_gpu, grid=(16*13,1,1), block=(128,1,1))
-			cuda.Context.synchronize()
-
-			mod_update = SourceModule(parameters + kernels.code_header + kernels.code_update, no_extern_c=True)
-			self.update = mod_update.get_function("update")
-			self.update.prepare("PPPPPPPPi")
-
-			self.encoded_X_training_gpu = cuda.mem_alloc(int(number_of_examples * self.number_of_patches * self.number_of_ta_chunks*4))
-			self.encode_X(X, self.encoded_X_training_gpu)
-
-			encoded_Y = np.empty((Y.shape[0], self.number_of_classes), dtype = np.int32)
-			for i in range(self.number_of_classes):
-				encoded_Y[:,i] = np.where(Y == i, self.T, -self.T)
-			self.Y_gpu = cuda.mem_alloc(encoded_Y.nbytes)
-			cuda.memcpy_htod(self.Y_gpu, encoded_Y)
-		elif incremental == False:
-			self.prepare(self.ta_state_gpu, self.clause_weights_gpu, self.clause_output_gpu, self.class_sum_gpu, grid=(16*13,1,1), block=(128,1,1))
-			cuda.Context.synchronize()
-
-		for epoch in range(epochs):
-			for e in range(0, number_of_examples, batch_size):
-				self.update.prepared_call((16*13,1,1), (128,1,1), g.state, self.ta_state_gpu, self.clause_weights_gpu, self.class_sum_gpu, self.clause_output_gpu, self.clause_patch_gpu, self.encoded_X_training_gpu, self.Y_gpu, np.int32(e))
-				cuda.Context.synchronize()
-
-		self.ta_state = np.array([])
-		self.clause_weights = np.array([])
+		self.ctm.fit(X, Y, epochs = epochs, incremental = incremental, batch_size = batch_size)
 
 		return
 
 	def score(self, X):
-		number_of_examples = X.shape[0]
-		
-		if not np.array_equal(self.X_test, X):
-			self.X_test = X
-			Xm = np.ascontiguousarray(X.flatten()).astype(np.uint32)
-			X_test_gpu = cuda.mem_alloc(Xm.nbytes)
-			cuda.memcpy_htod(X_test_gpu, Xm)
-			
-			self.encoded_X_test_gpu = cuda.mem_alloc(int(number_of_examples * self.number_of_patches * self.number_of_ta_chunks*4))
-			self.encode_X(X, self.encoded_X_test_gpu)
-
-			parameters = """
-#define CLASSES %d
-#define CLAUSES %d
-#define FEATURES %d
-#define STATE_BITS %d
-#define BOOST_TRUE_POSITIVE_FEEDBACK %d
-#define S %f
-#define THRESHOLD %d
-
-#define NEGATIVE_CLAUSES %d
-
-#define PATCHES %d
-
-#define NUMBER_OF_EXAMPLES %d
-
-#define BATCH_SIZE %d
-
-
-""" % (self.number_of_classes, self.number_of_clauses, self.number_of_features, self.number_of_state_bits, self.boost_true_positive_feedback, self.s, self.T, 1, 1, number_of_examples, 100)
-
-			mod = SourceModule(parameters + kernels.code_header + kernels.code_evaluate, no_extern_c=True)
-			self.evaluate = mod.get_function("evaluate")
-
-		class_sum = np.ascontiguousarray(np.zeros(self.number_of_classes*number_of_examples)).astype(np.int32)
-		class_sum_gpu = cuda.mem_alloc(class_sum.nbytes)
-		cuda.memcpy_htod(class_sum_gpu, class_sum)
-
-		self.evaluate(self.ta_state_gpu, self.clause_weights_gpu, class_sum_gpu, self.encoded_X_test_gpu, grid=(16*13,1,1), block=(128,1,1))
-		cuda.Context.synchronize()
-		cuda.memcpy_dtoh(class_sum, class_sum_gpu)
-		
-		class_sum = np.clip(class_sum.reshape((self.number_of_classes, number_of_examples)), -self.T, self.T)
-
-		return class_sum
+		return self.ctm.score(X)
 
 	def predict(self, X):
-		return np.argmax(self.score(X), axis=0)
+		return self.ctm.predict(X)
 
 class TsetlinMachine(CommonTsetlinMachine):
-	def __init__(self, number_of_clauses, T, s, boost_true_positive_feedback=1, number_of_state_bits=8, append_negated=True, max_weight=1):
+	def __init__(self, number_of_clauses, T, s, boost_true_positive_feedback=1, number_of_state_bits=8, append_negated=True, max_weight=1, number_of_classes = None):
 		self.number_of_clauses = number_of_clauses
-		self.number_of_clause_chunks = (number_of_clauses-1)/32 + 1
-		self.number_of_state_bits = number_of_state_bits
 		self.T = int(T)
 		self.s = s
-		self.max_weight = min(max_weight,255)
 		self.boost_true_positive_feedback = boost_true_positive_feedback
-		
+		self.number_of_state_bits = number_of_state_bits
 		self.append_negated = append_negated
+		self.max_weight = max_weight
 
 		self.X_train = np.array([])
 		self.Y_train = np.array([])
-		self.X_test = np.array([])
-		self.ta_state = np.array([])
-		self.clause_weights = np.array([])
-
-		mod_encode = SourceModule(kernels.code_encode, no_extern_c=True)
-		self.prepare_encode = mod_encode.get_function("prepare_encode")
-		self.encode = mod_encode.get_function("encode")
 
 	def fit(self, X, Y, epochs=100, incremental=False, batch_size = 100):
-		number_of_examples = X.shape[0]
-		number_of_example_chunks = int((number_of_examples-1)/32 + 1)
-
 		if (not np.array_equal(self.X_train, X)) or (not np.array_equal(self.Y_train, Y)):
 			self.X_train = X
 			self.Y_train = Y
 
-			self.number_of_classes = 1
+			self.tm = MultiClassTsetlinMachine(self.number_of_clauses, self.T, self.s, boost_true_positive_feedback=self.boost_true_positive_feedback, number_of_state_bits=self.number_of_state_bits, append_negated=self.append_negated, max_weight=self.max_weight, number_of_classes=1)
 
-			self.min_y = None
-			self.max_y = None
-			
-			self.dim = (X.shape[1], 1, 1)
-			self.patch_dim = (X.shape[1], 1, 1)
-
-			if self.append_negated:
-				self.number_of_features = X.shape[1]*2
-			else:
-				self.number_of_features = X.shape[1]
-
-			self.number_of_patches = 1
-			self.number_of_ta_chunks = int((self.number_of_features-1)/32 + 1)
-			
-			parameters = """
-	#define CLASSES %d
-	#define CLAUSES %d
-	#define FEATURES %d
-	#define STATE_BITS %d
-	#define BOOST_TRUE_POSITIVE_FEEDBACK %d
-	#define S %f
-	#define THRESHOLD %d
-	#define MAX_WEIGHT %d
-
-	#define NEGATIVE_CLAUSES %d
-
-	#define PATCHES %d
-
-	#define NUMBER_OF_EXAMPLES %d
-
-	#define BATCH_SIZE %d
-
-""" % (self.number_of_classes, self.number_of_clauses, self.number_of_features, self.number_of_state_bits, int(self.boost_true_positive_feedback), self.s, self.T, self.max_weight, 1, 1, number_of_examples, batch_size)
-
-			mod_prepare = SourceModule(parameters + kernels.code_header + kernels.code_prepare, no_extern_c=True)
-			self.prepare = mod_prepare.get_function("prepare")
-
-			self.allocate_gpu_memory(number_of_examples)
-
-			self.prepare(self.ta_state_gpu, self.clause_weights_gpu, self.clause_output_gpu, self.class_sum_gpu, grid=(16*13,1,1), block=(128,1,1))
-			cuda.Context.synchronize()
-
-			mod_update = SourceModule(parameters + kernels.code_header + kernels.code_update, no_extern_c=True)
-			self.update = mod_update.get_function("update")
-			self.update.prepare("PPPPPPPPi")
-
-			self.encoded_X_training_gpu = cuda.mem_alloc(int(number_of_examples * self.number_of_patches * self.number_of_ta_chunks*4))
-			self.encode_X(X, self.encoded_X_training_gpu)
-
-			encoded_Y = np.where(Y == 1, self.T, -self.T).astype(np.int32)
-			
-			self.Y_gpu = cuda.mem_alloc(encoded_Y.nbytes)
-			cuda.memcpy_htod(self.Y_gpu, encoded_Y)
-		elif incremental == False:
-			self.prepare(self.ta_state_gpu, self.clause_weights_gpu, self.clause_output_gpu, self.class_sum_gpu, grid=(16*13,1,1), block=(128,1,1))
-			cuda.Context.synchronize()
-
-		for epoch in range(epochs):
-			for e in range(0, number_of_examples, batch_size):
-				self.update.prepared_call((16*13,1,1), (128,1,1), g.state, self.ta_state_gpu, self.clause_weights_gpu, self.class_sum_gpu, self.clause_output_gpu, self.clause_patch_gpu, self.encoded_X_training_gpu, self.Y_gpu, np.int32(e))
-				cuda.Context.synchronize()
-
-		self.ta_state = np.array([])
-		self.clause_weights = np.array([])
+		self.tm.fit(X, Y, epochs = epochs, incremental = incremental, batch_size = batch_size)
 
 		return
 
 	def score(self, X):
-		number_of_examples = X.shape[0]
-		
-		if not np.array_equal(self.X_test, X):
-			self.X_test = X
-			Xm = np.ascontiguousarray(X.flatten()).astype(np.uint32)
-			X_test_gpu = cuda.mem_alloc(Xm.nbytes)
-			cuda.memcpy_htod(X_test_gpu, Xm)
-			
-			self.encoded_X_test_gpu = cuda.mem_alloc(int(number_of_examples * self.number_of_patches * self.number_of_ta_chunks*4))
-			self.encode_X(X, self.encoded_X_test_gpu)
-
-			parameters = """
-#define CLASSES %d
-#define CLAUSES %d
-#define FEATURES %d
-#define STATE_BITS %d
-#define BOOST_TRUE_POSITIVE_FEEDBACK %d
-#define S %f
-#define THRESHOLD %d
-
-#define NEGATIVE_CLAUSES %d
-
-#define PATCHES %d
-
-#define NUMBER_OF_EXAMPLES %d
-
-#define BATCH_SIZE %d
-
-
-""" % (self.number_of_classes, self.number_of_clauses, self.number_of_features, self.number_of_state_bits, self.boost_true_positive_feedback, self.s, self.T, 1, 1, number_of_examples, 100)
-
-			mod = SourceModule(parameters + kernels.code_header + kernels.code_evaluate, no_extern_c=True)
-			self.evaluate = mod.get_function("evaluate")
-
-		class_sum = np.ascontiguousarray(np.zeros(self.number_of_classes*number_of_examples)).astype(np.int32)
-		class_sum_gpu = cuda.mem_alloc(class_sum.nbytes)
-		cuda.memcpy_htod(class_sum_gpu, class_sum)
-
-		self.evaluate(self.ta_state_gpu, self.clause_weights_gpu, class_sum_gpu, self.encoded_X_test_gpu, grid=(16*13,1,1), block=(128,1,1))
-		cuda.Context.synchronize()
-		cuda.memcpy_dtoh(class_sum, class_sum_gpu)
-		
-		class_sum = np.clip(class_sum, -self.T, self.T)
-
-		return class_sum
+		return -1*self.tm.score(X)[0,:]
 
 	def predict(self, X):
-		return (self.score(X) >= 0)
+		return self.score(X) >= 0
 
 class RegressionTsetlinMachine(CommonTsetlinMachine):
 	def __init__(self, number_of_clauses, T, s, boost_true_positive_feedback=1, number_of_state_bits=8, append_negated=True, max_weight=1):
